from datetime import datetime
import xml.dom.minidom
import xml.parsers.expat
import logging

from . import errors
from . import xmlbuilder


logger = logging.getLogger(__name__)


def OsmResponseToDom(response, tag, single=False, allow_empty=False):
    """
    Returns the (sub-) DOM parsed from an OSM response
    """
    try:
        dom = xml.dom.minidom.parseString(response)
        osm_dom = dom.getElementsByTagName("osm")[0]
        all_data = osm_dom.getElementsByTagName(tag)
        first_element = all_data[0]
    except IndexError as e:
        if allow_empty:
            return []
        raise errors.XmlResponseInvalidError(
            f"The XML response from the OSM API is invalid: {e!r}"
        )
    except xml.parsers.expat.ExpatError as e:
        raise errors.XmlResponseInvalidError(
            f"The XML response from the OSM API is invalid: {e!r}"
        )

    if single:
        return first_element
    return all_data


def DomParseNode(DomElement):
    """
    Returns NodeData for the node.
    """
    result = _DomGetAttributes(DomElement)
    result["tag"] = _DomGetTag(DomElement)
    return result


def DomParseWay(DomElement):
    """
    Returns WayData for the way.
    """
    result = _DomGetAttributes(DomElement)
    result["tag"] = _DomGetTag(DomElement)
    result["nd"] = _DomGetNd(DomElement)
    return result


def DomParseRelation(DomElement):
    """
    Returns RelationData for the relation.
    """
    result = _DomGetAttributes(DomElement)
    result["tag"] = _DomGetTag(DomElement)
    result["member"] = _DomGetMember(DomElement)
    return result


def DomParseChangeset(DomElement, include_discussion=False):
    """
    Returns ChangesetData for the changeset.
    """
    result = _DomGetAttributes(DomElement)
    result["tag"] = _DomGetTag(DomElement)
    if include_discussion:
        result["discussion"] = _DomGetDiscussion(DomElement)

    return result


def DomParseNote(DomElement):
    """
    Returns NoteData for the note.
    """
    result = _DomGetAttributes(DomElement)
    result["id"] = xmlbuilder._GetXmlValue(DomElement, "id")
    result["status"] = xmlbuilder._GetXmlValue(DomElement, "status")

    result["date_created"] = _ParseDate(
        xmlbuilder._GetXmlValue(DomElement, "date_created")
    )
    result["date_closed"] = _ParseDate(
        xmlbuilder._GetXmlValue(DomElement, "date_closed")
    )
    result["comments"] = _DomGetComments(DomElement)

    return result


def _DomGetAttributes(DomElement):
    """
    Returns a formated dictionnary of attributes of a DomElement.
    """

    def is_true(v):
        return v == "true"

    attribute_mapping = {
        "uid": int,
        "changeset": int,
        "version": int,
        "id": int,
        "lat": float,
        "lon": float,
        "open": is_true,
        "visible": is_true,
        "ref": int,
        "comments_count": int,
        "timestamp": _ParseDate,
        "created_at": _ParseDate,
        "closed_at": _ParseDate,
        "date": _ParseDate,
    }
    result = {}
    for k, v in DomElement.attributes.items():
        try:
            result[k] = attribute_mapping[k](v)
        except KeyError:
            result[k] = v
    return result


def _DomGetTag(DomElement):
    """
    Returns the dictionnary of tags of a DomElement.
    """
    result = {}
    for t in DomElement.getElementsByTagName("tag"):
        k = t.attributes["k"].value
        v = t.attributes["v"].value
        result[k] = v
    return result


def _DomGetNd(DomElement):
    """
    Returns the list of nodes of a DomElement.
    """
    result = []
    for t in DomElement.getElementsByTagName("nd"):
        result.append(int(int(t.attributes["ref"].value)))
    return result


def _DomGetDiscussion(DomElement):
    """
    Returns the dictionnary of comments of a DomElement.
    """
    result = []
    try:
        discussion = DomElement.getElementsByTagName("discussion")[0]
        for t in discussion.getElementsByTagName("comment"):
            comment = _DomGetAttributes(t)
            comment["text"] = xmlbuilder._GetXmlValue(t, "text")
            result.append(comment)
    except IndexError:
        pass
    return result


def _DomGetComments(DomElement):
    """
    Returns the list of comments of a DomElement.
    """
    result = []
    for t in DomElement.getElementsByTagName("comment"):
        comment = {}
        comment["date"] = _ParseDate(xmlbuilder._GetXmlValue(t, "date"))
        comment["action"] = xmlbuilder._GetXmlValue(t, "action")
        comment["text"] = xmlbuilder._GetXmlValue(t, "text")
        comment["html"] = xmlbuilder._GetXmlValue(t, "html")
        comment["uid"] = xmlbuilder._GetXmlValue(t, "uid")
        comment["user"] = xmlbuilder._GetXmlValue(t, "user")
        result.append(comment)
    return result


def _DomGetMember(DomElement):
    """
    Returns a list of relation members.
    """
    result = []
    for m in DomElement.getElementsByTagName("member"):
        result.append(_DomGetAttributes(m))
    return result


def _ParseDate(DateString):
    date_formats = ["%Y-%m-%d %H:%M:%S UTC", "%Y-%m-%dT%H:%M:%SZ"]
    for date_format in date_formats:
        try:
            result = datetime.strptime(DateString, date_format)
            return result
        except (ValueError, TypeError):
            logger.debug(f"{DateString} does not match {date_format}")

    return DateString



import datetime
import itertools as it
import logging
import requests
import time

from . import errors


logger = logging.getLogger(__name__)


class OsmApiSession:
    MAX_RETRY_LIMIT = 5
    """Maximum retries if a call to the remote API fails (default: 5)"""

    def __init__(self, base_url, created_by, auth=None, session=None, timeout=30):
        self._api = base_url
        self._created_by = created_by
        self._timeout = timeout

        try:
            self._auth = auth
            if not auth and session.auth:
                self._auth = session.auth
        except AttributeError:
            pass

        self._http_session = session
        self._session = self._get_http_session()

    def close(self):
        if self._session:
            self._session.close()

    def _http_request(self, method, path, auth, send, return_value=True):  # noqa
        """
        Returns the response generated by an HTTP request.

        `method` is a HTTP method to be executed
        with the request data. For example: 'GET' or 'POST'.
        `path` is the path to the requested resource relative to the
        base API address stored in self._api. Should start with a
        slash character to separate the URL.
        `auth` is a boolean indicating whether authentication should
        be preformed on this request.
        `send` contains additional data that might be sent in a
        request.
        `return_value` indicates wheter this request should return
        any data or not.

        If the username or password is missing,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If the requested element has been deleted,
        `OsmApi.ElementDeletedApiError` is raised.

        If the requested element can not be found,
        `OsmApi.ElementNotFoundApiError` is raised.

        If the response status code indicates an error,
        `OsmApi.ApiError` is raised.
        """
        logger.debug(f"{datetime.datetime.now():%Y-%m-%d %H:%M:%S} {method} {path}")

        # Add API base URL to path
        path = self._api + path

        if auth and not self._auth:
            raise errors.UsernamePasswordMissingError("Username/Password missing")

        try:
            response = self._session.request(
                method, path, data=send, timeout=self._timeout
            )
        except requests.exceptions.Timeout as e:
            raise errors.TimeoutApiError(
                0, f"Request timed out (timeout={self._timeout})", ""
            ) from e
        except requests.exceptions.ConnectionError as e:
            raise errors.ConnectionApiError(0, f"Connection error: {str(e)}", "") from e
        except requests.exceptions.RequestException as e:
            raise errors.ApiError(0, str(e), "") from e

        if response.status_code != 200:
            payload = response.content.strip()
            if response.status_code == 401:
                raise errors.UnauthorizedApiError(
                    response.status_code, response.reason, payload
                )
            if response.status_code == 404:
                raise errors.ElementNotFoundApiError(
                    response.status_code, response.reason, payload
                )
            elif response.status_code == 410:
                raise errors.ElementDeletedApiError(
                    response.status_code, response.reason, payload
                )
            raise errors.ApiError(response.status_code, response.reason, payload)
        if return_value and not response.content:
            raise errors.ResponseEmptyApiError(
                response.status_code, response.reason, ""
            )

        logger.debug(f"{datetime.datetime.now():%Y-%m-%d %H:%M:%S} {method} {path}")
        return response.content

    def _http(self, cmd, path, auth, send, return_value=True):  # noqa
        for i in it.count(1):
            try:
                return self._http_request(
                    cmd, path, auth, send, return_value=return_value
                )
            except errors.ApiError as e:
                if e.status >= 500:
                    if i == self.MAX_RETRY_LIMIT:
                        raise
                    if i != 1:
                        self._sleep()
                    self._session = self._get_http_session()
                else:
                    logger.exception("ApiError Exception occured")
                    raise
            except errors.UsernamePasswordMissingError:
                raise
            except Exception as e:
                logger.exception("General exception occured")
                if i == self.MAX_RETRY_LIMIT:
                    if isinstance(e, errors.OsmApiError):
                        raise
                    raise errors.MaximumRetryLimitReachedError(
                        f"Give up after {i} retries"
                    ) from e
                if i != 1:
                    self._sleep()
                self._session = self._get_http_session()

    def _get_http_session(self):
        """
        Creates a requests session for connection pooling.
        """
        if self._http_session:
            session = self._http_session
        else:
            session = requests.Session()

        session.auth = self._auth
        session.headers.update({"user-agent": self._created_by})
        return session

    def _sleep(self):
        time.sleep(5)

    def _get(self, path):
        return self._http("GET", path, False, None)

    def _put(self, path, data, return_value=True):
        return self._http("PUT", path, True, data, return_value=return_value)

    def _post(self, path, data, optionalAuth=False, forceAuth=False):
        # the Notes API allows certain POSTs by non-authenticated users
        auth = optionalAuth and self._auth
        if forceAuth:
            auth = True
        return self._http("POST", path, auth, data)

    def _delete(self, path, data):
        return self._http("DELETE", path, True, data)



from . import osmapi_test
from datetime import datetime
import osmapi
from urllib import parse as urlparse


class TestOsmApiNotes(osmapi_test.TestOsmApi):
    def test_NotesGet(self):
        self._session_mock()

        result = self.api.NotesGet(-1.4998534, 45.9667901, -1.4831815, 52.4710193)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        urlParts = urlparse.urlparse(args[1])
        params = urlparse.parse_qs(urlParts.query)
        self.assertEqual(params["bbox"][0], "-1.499853,45.966790,-1.483181,52.471019")
        self.assertEqual(params["limit"][0], "100")
        self.assertEqual(params["closed"][0], "7")

        self.assertEqual(len(result), 14)
        self.assertEqual(
            result[2],
            {
                "id": "231775",
                "lon": -1.4929605,
                "lat": 52.4107312,
                "date_created": datetime(2014, 8, 28, 19, 25, 37),
                "date_closed": datetime(2014, 9, 27, 9, 21, 41),
                "status": "closed",
                "comments": [
                    {
                        "date": datetime(2014, 8, 28, 19, 25, 37),
                        "action": "opened",
                        "text": "Is it Paynes or Payne's",
                        "html": "<p>Is it Paynes or Payne's</p>",
                        "uid": "1486336",
                        "user": "Wyken Seagrave",
                    },
                    {
                        "date": datetime(2014, 9, 26, 13, 5, 33),
                        "action": "commented",
                        "text": "Royal Mail's postcode finder has PAYNES LANE",
                        "html": (
                            "<p>Royal Mail's postcode finder " "has PAYNES LANE</p>"
                        ),
                        "uid": None,
                        "user": None,
                    },
                ],
            },
        )

    def test_NotesGet_empty(self):
        self._session_mock()

        result = self.api.NotesGet(
            -93.8472901, 35.9763601, -80, 36.176360100000004, limit=1, closed=0
        )

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        urlParts = urlparse.urlparse(args[1])
        params = urlparse.parse_qs(urlParts.query)

        self.assertEqual(params["limit"][0], "1")
        self.assertEqual(params["closed"][0], "0")

        self.assertEqual(len(result), 0)
        self.assertEqual(result, [])

    def test_NoteGet(self):
        self._session_mock()

        result = self.api.NoteGet(1111)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], self.api_base + "/api/0.6/notes/1111")

        self.assertEqual(
            result,
            {
                "id": "1111",
                "lon": 12.3133135,
                "lat": 37.9305489,
                "date_created": datetime(2013, 5, 1, 20, 58, 21),
                "date_closed": datetime(2013, 8, 21, 16, 43, 26),
                "status": "closed",
                "comments": [
                    {
                        "date": datetime(2013, 5, 1, 20, 58, 21),
                        "action": "opened",
                        "text": "It does not exist this path",
                        "html": "<p>It does not exist this path</p>",
                        "uid": "1363438",
                        "user": "giuseppemari",
                    },
                    {
                        "date": datetime(2013, 8, 21, 16, 43, 26),
                        "action": "closed",
                        "text": "there is no path signed",
                        "html": "<p>there is no path signed</p>",
                        "uid": "1714220",
                        "user": "luschi",
                    },
                ],
            },
        )

    def test_NoteGet_invalid_xml(self):
        self._session_mock()

        with self.assertRaises(osmapi.XmlResponseInvalidError):
            self.api.NoteGet(1111)

    def test_NoteCreate(self):
        self._session_mock(auth=True)

        note = {"lat": 47.123, "lon": 8.432, "text": "This is a test"}
        result = self.api.NoteCreate(note)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "POST")
        urlParts = urlparse.urlparse(args[1])
        params = urlparse.parse_qs(urlParts.query)
        self.assertEqual(params["lat"][0], "47.123")
        self.assertEqual(params["lon"][0], "8.432")
        self.assertEqual(params["text"][0], "This is a test")

        self.assertEqual(
            result,
            {
                "id": "816",
                "lat": 47.123,
                "lon": 8.432,
                "date_created": datetime(2014, 10, 3, 15, 21, 21),
                "date_closed": None,
                "status": "open",
                "comments": [
                    {
                        "date": datetime(2014, 10, 3, 15, 21, 22),
                        "action": "opened",
                        "text": "This is a test",
                        "html": "<p>This is a test</p>",
                        "uid": "1841",
                        "user": "metaodi",
                    }
                ],
            },
        )

    def test_NoteCreateAnonymous(self):
        self._session_mock()

        note = {"lat": 47.123, "lon": 8.432, "text": "test 123"}
        result = self.api.NoteCreate(note)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "POST")
        urlParts = urlparse.urlparse(args[1])
        params = urlparse.parse_qs(urlParts.query)
        self.assertEqual(params["lat"][0], "47.123")
        self.assertEqual(params["lon"][0], "8.432")
        self.assertEqual(params["text"][0], "test 123")

        self.assertEqual(
            result,
            {
                "id": "842",
                "lat": 58.3368222,
                "lon": 25.8826183,
                "date_created": datetime(2015, 1, 3, 10, 49, 39),
                "date_closed": None,
                "status": "open",
                "comments": [
                    {
                        "date": datetime(2015, 1, 3, 10, 49, 39),
                        "action": "opened",
                        "text": "test 123",
                        "html": "<p>test 123</p>",
                        "uid": None,
                        "user": None,
                    }
                ],
            },
        )

    def test_NoteComment(self):
        self._session_mock(auth=True)

        result = self.api.NoteComment(812, "This is a comment")

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "POST")
        self.assertEqual(
            args[1], self.api_base + "/api/0.6/notes/812/comment?text=This+is+a+comment"
        )

        self.assertEqual(
            result,
            {
                "id": "812",
                "lat": 47.123,
                "lon": 8.432,
                "date_created": datetime(2014, 10, 3, 15, 11, 5),
                "date_closed": None,
                "status": "open",
                "comments": [
                    {
                        "date": datetime(2014, 10, 3, 15, 11, 5),
                        "action": "opened",
                        "text": "This is a test",
                        "html": "<p>This is a test</p>",
                        "uid": "1841",
                        "user": "metaodi",
                    },
                    {
                        "date": datetime(2014, 10, 4, 22, 36, 35),
                        "action": "commented",
                        "text": "This is a comment",
                        "html": "<p>This is a comment</p>",
                        "uid": "1841",
                        "user": "metaodi",
                    },
                ],
            },
        )

    def test_NoteCommentAnonymous(self):
        self._session_mock()

        result = self.api.NoteComment(842, "blubb")

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "POST")
        self.assertEqual(
            args[1], self.api_base + "/api/0.6/notes/842/comment?text=blubb"
        )

        self.assertEqual(
            result,
            {
                "id": "842",
                "lat": 58.3368222,
                "lon": 25.8826183,
                "date_created": datetime(2015, 1, 3, 10, 49, 39),
                "date_closed": None,
                "status": "open",
                "comments": [
                    {
                        "date": datetime(2015, 1, 3, 10, 49, 39),
                        "action": "opened",
                        "text": "test 123",
                        "html": "<p>test 123</p>",
                        "uid": None,
                        "user": None,
                    },
                    {
                        "date": datetime(2015, 1, 3, 11, 6, 0),
                        "action": "commented",
                        "text": "blubb",
                        "html": "<p>blubb</p>",
                        "uid": None,
                        "user": None,
                    },
                ],
            },
        )

    def test_NoteCommentOnClosedNote(self):
        self._session_mock(status=409)

        with self.assertRaises(osmapi.NoteAlreadyClosedApiError) as cm:
            self.api.NoteComment(817, "Comment on closed note")

        self.assertEqual(cm.exception.status, 409)
        self.assertEqual(
            cm.exception.payload, "The note 817 was closed at 2022-04-29 20:57:20 UTC"
        )

    def test_NoteComment_non_existing_note(self):
        self._session_mock(status=404)

        with self.assertRaises(osmapi.ElementNotFoundApiError) as cm:
            self.api.NoteComment(817, "Comment on closed note")

        self.assertEqual(cm.exception.status, 404)

    def test_NoteClose(self):
        self._session_mock(auth=True)

        result = self.api.NoteClose(819, "Close this note!")

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "POST")
        self.assertEqual(
            args[1], self.api_base + "/api/0.6/notes/819/close?text=Close+this+note%21"
        )

        self.assertEqual(
            result,
            {
                "id": "815",
                "lat": 47.123,
                "lon": 8.432,
                "date_created": datetime(2014, 10, 3, 15, 20, 57),
                "date_closed": datetime(2014, 10, 5, 16, 35, 13),
                "status": "closed",
                "comments": [
                    {
                        "date": datetime(2014, 10, 3, 15, 20, 57),
                        "action": "opened",
                        "text": "This is a test",
                        "html": "<p>This is a test</p>",
                        "uid": "1841",
                        "user": "metaodi",
                    },
                    {
                        "date": datetime(2014, 10, 5, 16, 35, 13),
                        "action": "closed",
                        "text": "Close this note!",
                        "html": "<p>Close this note!</p>",
                        "uid": "1841",
                        "user": "metaodi",
                    },
                ],
            },
        )

    def test_NoteAlreadyClosed(self):
        self._session_mock(auth=True, status=409)

        with self.assertRaises(osmapi.NoteAlreadyClosedApiError) as cm:
            self.api.NoteClose(819, "Close this note!")

        self.assertEqual(cm.exception.status, 409)
        self.assertEqual(
            cm.exception.payload, "The note 819 was closed at 2022-04-29 20:57:20 UTC"
        )

    def test_NoteReopen(self):
        self._session_mock(auth=True)

        result = self.api.NoteReopen(815, "Reopen this note!")

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "POST")
        self.assertEqual(
            args[1],
            (self.api_base + "/api/0.6/notes/815/reopen?text=Reopen+this+note%21"),
        )

        self.assertEqual(
            result,
            {
                "id": "815",
                "lat": 47.123,
                "lon": 8.432,
                "date_created": datetime(2014, 10, 3, 15, 20, 57),
                "date_closed": None,
                "status": "open",
                "comments": [
                    {
                        "date": datetime(2014, 10, 3, 15, 20, 57),
                        "action": "opened",
                        "text": "This is a test",
                        "html": "<p>This is a test</p>",
                        "uid": "1841",
                        "user": "metaodi",
                    },
                    {
                        "date": datetime(2014, 10, 5, 16, 35, 13),
                        "action": "closed",
                        "text": "Close this note!",
                        "html": "<p>Close this note!</p>",
                        "uid": "1841",
                        "user": "metaodi",
                    },
                    {
                        "date": datetime(2014, 10, 5, 16, 44, 56),
                        "action": "reopened",
                        "text": "Reopen this note!",
                        "html": "<p>Reopen this note!</p>",
                        "uid": "1841",
                        "user": "metaodi",
                    },
                ],
            },
        )

    def test_NotesSearch(self):
        self._session_mock()

        result = self.api.NotesSearch("street")

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        urlParts = urlparse.urlparse(args[1])
        params = urlparse.parse_qs(urlParts.query)
        self.assertEqual(params["q"][0], "street")
        self.assertEqual(params["limit"][0], "100")
        self.assertEqual(params["closed"][0], "7")

        self.assertEqual(len(result), 3)
        self.assertEqual(
            result[1],
            {
                "id": "788",
                "lon": 11.96395,
                "lat": 57.70301,
                "date_created": datetime(2014, 7, 16, 16, 12, 41),
                "date_closed": None,
                "status": "open",
                "comments": [
                    {
                        "date": datetime(2014, 7, 16, 16, 12, 41),
                        "action": "opened",
                        "text": "One way street:\ncomment",
                        "html": "<p>One way street:\n<br />comment</p>",
                        "uid": None,
                        "user": None,
                    }
                ],
            },
        )



# This script shows how to authenticate with OAuth2 with a backend application
# The token is saved to disk in $HOME/.osmapi/token.json
# It can be reused until it's revoked or expired.

# install oauthlib for requests:  pip install oauthlib requests-oauthlib
from requests_oauthlib import OAuth2Session
import json
import webbrowser
import osmapi
from dotenv import load_dotenv, find_dotenv
import os
import sys

load_dotenv(find_dotenv())

# Credentials you get from registering a new application
# register here: https://master.apis.dev.openstreetmap.org/oauth2/applications
# or on production: https://www.openstreetmap.org/oauth2/applications
client_id = os.getenv("OSM_OAUTH_CLIENT_ID")
client_secret = os.getenv("OSM_OAUTH_CLIENT_SECRET")

if client_id is None:
    raise
if client_secret is None:
    raise

# special value for redirect_uri for non-web applications
redirect_uri = "urn:ietf:wg:oauth:2.0:oob"

authorization_base_url = "https://master.apis.dev.openstreetmap.org/oauth2/authorize"
token_url = "https://master.apis.dev.openstreetmap.org/oauth2/token"
scope = ["write_api", "write_notes"]


def get_osmapi_path():
    base_dir = ""

    if os.getenv("HOME"):
        base_dir = os.getenv("HOME")
    elif os.getenv("HOMEDRIVE") and os.getenv("HOMEPATH"):
        base_dir = os.path.join(os.getenv("HOMEDRIVE"), os.getenv("HOMEPATH"))
    elif os.getenv("USERPROFILE"):
        base_dir = os.getenv("USERPROFILE")

    if not base_dir:
        print(
            "Unable to find home directory (check env vars HOME, HOMEDRIVE, HOMEPATH and USERPROFILE)",  # noqa
            file=sys.stderr,
        )
        raise Exception("Home directory not found")

    return os.path.join(base_dir, ".osmapi")


def token_saver(token):
    osmapi_path = get_osmapi_path()
    token_path = os.path.join(osmapi_path, "token.json")

    with open(token_path, "w") as f:
        print(f"Saving token {token} to {token_path}")
        f.write(json.dumps(token))


def token_loader():
    osmapi_path = get_osmapi_path()
    token_path = os.path.join(osmapi_path, "token.json")

    with open(token_path, "r") as f:
        token = json.loads(f.read())
        print(f"Loaded token {token} from {token_path}")
    return token


def save_and_get_access_token(client_id, client_secret, redirect_uri, scope):
    oauth = OAuth2Session(
        client_id=client_id,
        redirect_uri=redirect_uri,
        scope=scope,
    )

    login_url, _ = oauth.authorization_url(authorization_base_url)

    print(f"Authorize user using this URL: {login_url}")
    webbrowser.open(login_url)

    authorization_code = input("Paste the authorization code here: ")

    token = oauth.fetch_token(
        token_url=token_url,
        client_secret=client_secret,
        code=authorization_code,
    )

    token_saver(token)
    return token


def make_osm_change(oauth_session):
    api = osmapi.OsmApi(
        api="https://api06.dev.openstreetmap.org", session=oauth_session
    )
    with api.Changeset({"comment": "My first test"}) as changeset_id:
        print(f"Part of Changeset {changeset_id}")
        node1 = api.NodeCreate({"lon": 1, "lat": 1, "tag": {}})
        print(node1)


# load a previously saved token
try:
    token = token_loader()
except FileNotFoundError:
    print("Token not found, get a new one...")
    token = save_and_get_access_token(client_id, client_secret, redirect_uri, scope)

# test the token
try:
    oauth_session = OAuth2Session(client_id, token=token)
    make_osm_change(oauth_session)
except osmapi.errors.UnauthorizedApiError:
    print("Token expired, let's create a new one")
    token = save_and_get_access_token(client_id, client_secret, redirect_uri, scope)
    oauth_session = OAuth2Session(client_id, token=token)
    make_osm_change(oauth_session)



from . import osmapi_test
import osmapi
from unittest import mock
import datetime


class TestOsmApiRelation(osmapi_test.TestOsmApi):
    def test_RelationGet(self):
        self._session_mock()

        result = self.api.RelationGet(321)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], self.api_base + "/api/0.6/relation/321")

        self.assertEqual(
            result,
            {
                "id": 321,
                "changeset": 434,
                "uid": 12,
                "timestamp": datetime.datetime(2009, 9, 15, 22, 24, 25),
                "visible": True,
                "version": 1,
                "user": "green525",
                "tag": {
                    "admin_level": "9",
                    "boundary": "administrative",
                    "type": "multipolygon",
                },
                "member": [
                    {"ref": 6908, "role": "outer", "type": "way"},
                    {"ref": 6352, "role": "outer", "type": "way"},
                    {"ref": 5669, "role": "outer", "type": "way"},
                    {"ref": 5682, "role": "outer", "type": "way"},
                    {"ref": 6909, "role": "outer", "type": "way"},
                    {"ref": 6355, "role": "outer", "type": "way"},
                    {"ref": 6910, "role": "outer", "type": "way"},
                    {"ref": 6911, "role": "outer", "type": "way"},
                    {"ref": 6912, "role": "outer", "type": "way"},
                ],
            },
        )

    def test_RelationGet_with_version(self):
        self._session_mock()

        result = self.api.RelationGet(765, 2)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], self.api_base + "/api/0.6/relation/765/2")

        self.assertEqual(result["id"], 765)
        self.assertEqual(result["changeset"], 41378)
        self.assertEqual(result["user"], "metaodi")
        self.assertEqual(result["tag"]["source"], "test")

    def test_RelationCreate(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=3333)
        self.api._CurrentChangesetId = 3333

        test_relation = {
            "tag": {
                "type": "test",
            },
            "member": [
                {"ref": 6908, "role": "outer", "type": "way"},
                {"ref": 6352, "role": "point", "type": "node"},
            ],
        }

        cs = self.api.ChangesetCreate({"comment": "This is a test relation"})
        self.assertEqual(cs, 3333)

        result = self.api.RelationCreate(test_relation)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "PUT")
        self.assertEqual(args[1], self.api_base + "/api/0.6/relation/create")

        self.assertEqual(result["id"], 8989)
        self.assertEqual(result["version"], 1)
        self.assertEqual(result["member"], test_relation["member"])
        self.assertEqual(result["tag"], test_relation["tag"])

    def test_RelationCreate_existing_node(self):
        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=1111)
        self.api._CurrentChangesetId = 1111

        test_relation = {
            "id": 456,
            "tag": {
                "type": "test",
            },
            "member": [
                {"ref": 6908, "role": "outer", "type": "way"},
                {"ref": 6352, "role": "point", "type": "node"},
            ],
        }

        with self.assertRaisesRegex(
            osmapi.OsmTypeAlreadyExistsError, "This relation already exists"
        ):
            self.api.RelationCreate(test_relation)

    def test_RelationUpdate(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=3333)
        self.api._CurrentChangesetId = 3333

        test_relation = {
            "id": 8989,
            "tag": {
                "type": "test update",
            },
            "member": [{"ref": 6908, "role": "outer", "type": "way"}],
        }

        cs = self.api.ChangesetCreate({"comment": "This is a test relation"})
        self.assertEqual(cs, 3333)

        result = self.api.RelationUpdate(test_relation)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "PUT")
        self.assertEqual(args[1], self.api_base + "/api/0.6/relation/8989")

        self.assertEqual(result["id"], 8989)
        self.assertEqual(result["version"], 42)
        self.assertEqual(result["member"], test_relation["member"])
        self.assertEqual(result["tag"], test_relation["tag"])

    def test_RelationDelete(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=3333)
        self.api._CurrentChangesetId = 3333

        test_relation = {"id": 8989}

        cs = self.api.ChangesetCreate({"comment": "This is a test relation"})
        self.assertEqual(cs, 3333)

        result = self.api.RelationDelete(test_relation)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "DELETE")
        self.assertEqual(args[1], self.api_base + "/api/0.6/relation/8989")

        self.assertEqual(result["id"], 8989)
        self.assertEqual(result["version"], 43)

    def test_RelationHistory(self):
        self._session_mock()

        result = self.api.RelationHistory(2470397)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], f"{self.api_base}/api/0.6/relation/2470397/history")

        self.assertEqual(len(result), 2)
        self.assertEqual(result[1]["id"], 2470397)
        self.assertEqual(result[1]["version"], 1)
        self.assertEqual(
            result[1]["tag"],
            {
                "restriction": "only_straight_on",
                "type": "restriction",
            },
        )
        self.assertEqual(result[2]["version"], 2)

    def test_RelationRelations(self):
        self._session_mock()

        result = self.api.RelationRelations(1532552)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], f"{self.api_base}/api/0.6/relation/1532552/relations")

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]["id"], 1532553)
        self.assertEqual(result[0]["version"], 85)
        self.assertEqual(len(result[0]["member"]), 120)
        self.assertEqual(result[0]["tag"]["type"], "network")
        self.assertEqual(result[0]["tag"]["name"], "Aargauischer Radroutennetz")

    def test_RelationRelationsUnusedElement(self):
        self._session_mock()

        result = self.api.RelationRelations(1532552)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], f"{self.api_base}/api/0.6/relation/1532552/relations")

        self.assertEqual(len(result), 0)
        self.assertIsInstance(result, list)

    def test_RelationFull(self):
        self._session_mock()

        result = self.api.RelationFull(2470397)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], f"{self.api_base}/api/0.6/relation/2470397/full")

        self.assertEqual(len(result), 11)
        self.assertEqual(result[1]["data"]["id"], 101142277)
        self.assertEqual(result[1]["data"]["version"], 8)
        self.assertEqual(result[1]["type"], "node")
        self.assertEqual(result[10]["data"]["id"], 2470397)
        self.assertEqual(result[10]["data"]["version"], 2)
        self.assertEqual(result[10]["type"], "relation")

    def test_RelationsGet(self):
        self._session_mock()

        result = self.api.RelationsGet([1532552, 1532553])

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(
            args[1], f"{self.api_base}/api/0.6/relations?relations=1532552,1532553"
        )

        self.assertEqual(len(result), 2)
        self.assertEqual(result[1532553]["id"], 1532553)
        self.assertEqual(result[1532553]["version"], 85)
        self.assertEqual(result[1532553]["user"], "SimonPoole")
        self.assertEqual(result[1532552]["id"], 1532552)
        self.assertEqual(result[1532552]["visible"], True)
        self.assertEqual(result[1532552]["tag"]["route"], "bicycle")

    def test_RelationFull_with_deleted_relation(self):
        self._session_mock(filenames=[], status=410)

        with self.assertRaises(osmapi.ElementDeletedApiError) as context:
            self.api.RelationFull(2911456)
        self.assertEqual(410, context.exception.status)



def _XmlBuild(ElementType, ElementData, WithHeaders=True, data=None):  # noqa
    xml = ""
    if WithHeaders:
        xml += '<?xml version="1.0" encoding="UTF-8"?>\n'
        xml += '<osm version="0.6" generator="'
        xml += data._created_by + '">\n'

    # <element attr="val">
    xml += "  <" + ElementType
    if "id" in ElementData:
        xml += ' id="' + str(ElementData["id"]) + '"'
    if "lat" in ElementData:
        xml += ' lat="' + str(ElementData["lat"]) + '"'
    if "lon" in ElementData:
        xml += ' lon="' + str(ElementData["lon"]) + '"'
    if "version" in ElementData:
        xml += ' version="' + str(ElementData["version"]) + '"'
    visible_str = str(ElementData.get("visible", True)).lower()
    xml += ' visible="' + visible_str + '"'
    if ElementType in ["node", "way", "relation"]:
        xml += ' changeset="' + str(data._CurrentChangesetId) + '"'
    xml += ">\n"

    # <tag... />
    for k, v in ElementData.get("tag", {}).items():
        xml += '    <tag k="' + _XmlEncode(k)
        xml += '" v="' + _XmlEncode(v) + '"/>\n'

    # <member... />
    for member in ElementData.get("member", []):
        xml += '    <member type="' + member["type"]
        xml += '" ref="' + str(member["ref"])
        xml += '" role="' + _XmlEncode(member["role"])
        xml += '"/>\n'

    # <nd... />
    for ref in ElementData.get("nd", []):
        xml += '    <nd ref="' + str(ref) + '"/>\n'

    # </element>
    xml += "  </" + ElementType + ">\n"

    if WithHeaders:
        xml += "</osm>\n"

    return xml.encode("utf8")


def _XmlEncode(text):
    return (
        text.replace("&", "&amp;")
        .replace('"', "&quot;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
    )


def _GetXmlValue(DomElement, tag):
    try:
        elem = DomElement.getElementsByTagName(tag)[0]
        return elem.firstChild.nodeValue
    except Exception:
        return None



# install oauthlib for requests:  pip install requests-oauth2client
from requests_oauth2client import OAuth2Client, OAuth2AuthorizationCodeAuth
import requests
import webbrowser
import osmapi
from dotenv import load_dotenv, find_dotenv
import os

load_dotenv(find_dotenv())

# Credentials you get from registering a new application
# register here: https://master.apis.dev.openstreetmap.org/oauth2/applications
# or on production: https://www.openstreetmap.org/oauth2/applications
client_id = os.getenv("OSM_OAUTH_CLIENT_ID")
client_secret = os.getenv("OSM_OAUTH_CLIENT_SECRET")

# special value for redirect_uri for non-web applications
redirect_uri = "urn:ietf:wg:oauth:2.0:oob"

authorization_base_url = "https://master.apis.dev.openstreetmap.org/oauth2/authorize"
token_url = "https://master.apis.dev.openstreetmap.org/oauth2/token"

oauth2client = OAuth2Client(
    token_endpoint=token_url,
    authorization_endpoint=authorization_base_url,
    redirect_uri=redirect_uri,
    auth=(client_id, client_secret),
    code_challenge_method=None,
)

# open OSM website to authrorize user using the write_api and write_notes scope
scope = ["write_api", "write_notes"]
az_request = oauth2client.authorization_request(scope=scope)
print(f"Authorize user using this URL: {az_request.uri}")
webbrowser.open(az_request.uri)

# create a new requests session using the OAuth authorization
auth_code = input("Paste the authorization code here: ")
auth = OAuth2AuthorizationCodeAuth(
    oauth2client,
    auth_code,
    redirect_uri=redirect_uri,
)
oauth_session = requests.Session()
oauth_session.auth = auth

# use the custom session
api = osmapi.OsmApi(api="https://api06.dev.openstreetmap.org", session=oauth_session)
with api.Changeset({"comment": "My first test"}) as changeset_id:
    print(f"Part of Changeset {changeset_id}")
    node1 = api.NodeCreate({"lon": 1, "lat": 1, "tag": {}})
    print(node1)



# install cli-oauth2 for requests:  pip install cli-oauth2
from oauthcli import OpenStreetMapDevAuth
import osmapi
from dotenv import load_dotenv, find_dotenv
import os

load_dotenv(find_dotenv())

# load secrets for OAuth
client_id = os.getenv("OSM_OAUTH_CLIENT_ID")
client_secret = os.getenv("OSM_OAUTH_CLIENT_SECRET")

auth = OpenStreetMapDevAuth(
    client_id, client_secret, ["write_api", "write_notes"]
).auth_code()


api = osmapi.OsmApi(api="https://api06.dev.openstreetmap.org", session=auth.session)
with api.Changeset({"comment": "My first test"}) as changeset_id:
    print(f"Part of Changeset {changeset_id}")
    node1 = api.NodeCreate({"lon": 1, "lat": 1, "tag": {}})
    print(node1)
    node2 = api.NodeCreate({"lon": 2, "lat": 2, "tag": {}})
    print(node2)
    way = api.WayCreate(
        {
            "nd": [
                node1["id"],
                node2["id"],
            ],
            "tag": {
                "highway": "unclassified",
                "name": "Osmapi Street",
            },
        }
    )
    print(way)



from . import osmapi_test
import osmapi
from unittest import mock
import datetime
from requests.auth import HTTPBasicAuth


class TestOsmApiNode(osmapi_test.TestOsmApi):
    def test_NodeGet(self):
        self._session_mock()

        result = self.api.NodeGet(123)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], self.api_base + "/api/0.6/node/123")

        self.assertEqual(
            result,
            {
                "id": 123,
                "changeset": 15293,
                "uid": 605,
                "timestamp": datetime.datetime(2012, 4, 18, 11, 14, 26),
                "lat": 51.8753146,
                "lon": -1.4857118,
                "visible": True,
                "version": 8,
                "user": "freundchen",
                "tag": {"amenity": "school", "foo": "bar", "name": "Berolina & Schule"},
            },
        )

    def test_NodeGet_with_version(self):
        self._session_mock()

        result = self.api.NodeGet(123, NodeVersion=2)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], self.api_base + "/api/0.6/node/123/2")

        self.assertEqual(
            result,
            {
                "id": 123,
                "changeset": 4152,
                "uid": 605,
                "timestamp": datetime.datetime(2011, 4, 18, 11, 14, 26),
                "lat": 51.8753146,
                "lon": -1.4857118,
                "visible": True,
                "version": 2,
                "user": "freundchen",
                "tag": {
                    "amenity": "school",
                },
            },
        )

    def test_NodeGet_invalid_response(self):
        self._session_mock()

        with self.assertRaises(osmapi.XmlResponseInvalidError):
            self.api.NodeGet(987)

    def test_NodeCreate_changesetauto(self):
        for filename in [
            "test_NodeCreate_changesetauto.xml",
            "test_ChangesetUpload_create_node.xml",
            "test_ChangesetClose.xml",
        ]:
            # setup mock
            self._session_mock(auth=True, filenames=[filename])
            self.api = osmapi.OsmApi(
                api="api06.dev.openstreetmap.org",
                changesetauto=True,
                session=self.session_mock,
            )
            self.api._session._sleep = mock.Mock()

            test_node = {
                "lat": 47.123,
                "lon": 8.555,
                "tag": {"amenity": "place_of_worship", "religion": "pastafarian"},
            }

            self.assertIsNone(self.api.NodeCreate(test_node))

    def test_NodeCreate(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=1111)
        self.api._CurrentChangesetId = 1111

        test_node = {
            "lat": 47.287,
            "lon": 8.765,
            "tag": {"amenity": "place_of_worship", "religion": "pastafarian"},
        }

        cs = self.api.ChangesetCreate({"comment": "This is a test dataset"})
        self.assertEqual(cs, 1111)
        result = self.api.NodeCreate(test_node)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "PUT")
        self.assertEqual(args[1], self.api_base + "/api/0.6/node/create")

        self.assertEqual(result["id"], 9876)
        self.assertEqual(result["lat"], test_node["lat"])
        self.assertEqual(result["lon"], test_node["lon"])
        self.assertEqual(result["tag"], test_node["tag"])

    def test_NodeCreate_wo_changeset(self):
        test_node = {
            "lat": 47.287,
            "lon": 8.765,
            "tag": {"amenity": "place_of_worship", "religion": "pastafarian"},
        }

        with self.assertRaisesRegex(
            osmapi.NoChangesetOpenError, "need to open a changeset"
        ):
            self.api.NodeCreate(test_node)

    def test_NodeCreate_existing_node(self):
        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=1111)
        self.api._CurrentChangesetId = 1111

        test_node = {
            "id": 123,
            "lat": 47.287,
            "lon": 8.765,
            "tag": {"amenity": "place_of_worship", "religion": "pastafarian"},
        }

        with self.assertRaisesRegex(
            osmapi.OsmTypeAlreadyExistsError, "This node already exists"
        ):
            self.api.NodeCreate(test_node)

    def test_NodeCreate_wo_auth(self):
        self._session_mock()

        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=1111)
        self.api._CurrentChangesetId = 1111
        test_node = {
            "lat": 47.287,
            "lon": 8.765,
            "tag": {"amenity": "place_of_worship", "religion": "pastafarian"},
        }

        with self.assertRaisesRegex(
            osmapi.UsernamePasswordMissingError, "Username/Password missing"
        ):
            self.api.NodeCreate(test_node)

    def test_NodeCreate_unauthorized(self):
        self._session_mock(auth=True, status=401)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=1111)
        self.api._CurrentChangesetId = 1111
        test_node = {
            "lat": 47.287,
            "lon": 8.765,
            "tag": {"amenity": "place_of_worship", "religion": "pastafarian"},
        }

        with self.assertRaises(osmapi.UnauthorizedApiError):
            self.api.NodeCreate(test_node)

    def test_NodeCreate_with_session_auth(self):
        self._session_mock()
        self.session_mock.auth = HTTPBasicAuth("user", "pass")

        api = osmapi.OsmApi(api=self.api_base, session=self.session_mock)

        # setup mock
        api.ChangesetCreate = mock.Mock(return_value=1111)
        api._CurrentChangesetId = 1111
        test_node = {
            "lat": 47.287,
            "lon": 8.765,
            "tag": {"amenity": "place_of_worship", "religion": "pastafarian"},
        }

        cs = api.ChangesetCreate({"comment": "This is a test dataset"})
        self.assertEqual(cs, 1111)
        result = api.NodeCreate(test_node)
        self.assertEqual(result["id"], 3322)

    def test_NodeCreate_with_exception(self):
        self._session_mock(auth=True)
        self.api._session._http_request = mock.Mock(side_effect=Exception)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=1111)
        self.api._CurrentChangesetId = 1111
        test_node = {
            "lat": 47.287,
            "lon": 8.765,
            "tag": {"amenity": "place_of_worship", "religion": "pastafarian"},
        }

        with self.assertRaisesRegex(
            osmapi.MaximumRetryLimitReachedError, "Give up after 5 retries"
        ):
            self.api.NodeCreate(test_node)

    def test_NodeUpdate(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=1111)
        self.api._CurrentChangesetId = 1111

        test_node = {
            "id": 7676,
            "lat": 47.287,
            "lon": 8.765,
            "tag": {"amenity": "place_of_worship", "name": "christian"},
        }

        cs = self.api.ChangesetCreate({"comment": "This is a test dataset"})
        self.assertEqual(cs, 1111)
        result = self.api.NodeUpdate(test_node)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "PUT")
        self.assertEqual(args[1], self.api_base + "/api/0.6/node/7676")

        self.assertEqual(result["id"], 7676)
        self.assertEqual(result["version"], 3)
        self.assertEqual(result["lat"], test_node["lat"])
        self.assertEqual(result["lon"], test_node["lon"])
        self.assertEqual(result["tag"], test_node["tag"])

    def test_NodeUpdateWhenChangesetIsClosed(self):
        self._session_mock(auth=True, status=409)

        self.api.ChangesetCreate = mock.Mock(return_value=1111)
        self.api._CurrentChangesetId = 1111

        test_node = {
            "id": 7676,
            "lat": 47.287,
            "lon": 8.765,
            "tag": {"amenity": "place_of_worship", "name": "christian"},
        }

        self.api.ChangesetCreate({"comment": "This is a test dataset"})

        with self.assertRaises(osmapi.ChangesetClosedApiError) as cm:
            self.api.NodeUpdate(test_node)

        self.assertEqual(cm.exception.status, 409)
        self.assertEqual(
            cm.exception.payload,
            "The changeset 2222 was closed at 2021-11-20 09:42:47 UTC.",
        )

    def test_NodeUpdateConflict(self):
        self._session_mock(auth=True, status=409)

        self.api.ChangesetCreate = mock.Mock(return_value=1111)
        self.api._CurrentChangesetId = 1111

        test_node = {
            "id": 7676,
            "lat": 47.287,
            "lon": 8.765,
            "tag": {"amenity": "place_of_worship", "name": "christian"},
        }

        self.api.ChangesetCreate({"comment": "This is a test dataset"})

        with self.assertRaises(osmapi.VersionMismatchApiError) as cm:
            self.api.NodeUpdate(test_node)

        self.assertEqual(cm.exception.status, 409)
        self.assertEqual(
            cm.exception.payload,
            "Version does not match the current database version of the element",
        )

    def test_NodeDelete(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=1111)
        self.api._CurrentChangesetId = 1111

        test_node = {"id": 7676}

        cs = self.api.ChangesetCreate({"comment": "This is a test dataset"})
        self.assertEqual(cs, 1111)

        result = self.api.NodeDelete(test_node)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "DELETE")
        self.assertEqual(args[1], self.api_base + "/api/0.6/node/7676")
        self.assertEqual(result["id"], 7676)
        self.assertEqual(result["version"], 4)

    def test_NodeHistory(self):
        self._session_mock()

        result = self.api.NodeHistory(123)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], self.api_base + "/api/0.6/node/123/history")

        self.assertEqual(len(result), 8)
        self.assertEqual(result[4]["id"], 123)
        self.assertEqual(result[4]["version"], 4)
        self.assertEqual(result[4]["lat"], 51.8753146)
        self.assertEqual(result[4]["lon"], -1.4857118)
        self.assertEqual(
            result[4]["tag"],
            {
                "empty": "",
                "foo": "bar",
            },
        )

    def test_NodeWays(self):
        self._session_mock()

        result = self.api.NodeWays(234)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], self.api_base + "/api/0.6/node/234/ways")

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]["id"], 60)
        self.assertEqual(result[0]["changeset"], 61)
        self.assertEqual(
            result[0]["tag"],
            {
                "highway": "path",
                "name": "Dog walking path",
            },
        )

    def test_NodeWaysNotExists(self):
        self._session_mock()

        result = self.api.NodeWays(404)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], f"{self.api_base}/api/0.6/node/404/ways")

        self.assertEqual(len(result), 0)
        self.assertIsInstance(result, list)

    def test_NodeRelations(self):
        self._session_mock()

        result = self.api.NodeRelations(4295668179)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], f"{self.api_base}/api/0.6/node/4295668179/relations")

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]["id"], 4294968148)
        self.assertEqual(result[0]["changeset"], 23123)
        self.assertEqual(
            result[0]["member"][1],
            {
                "role": "point",
                "ref": 4295668179,
                "type": "node",
            },
        )
        self.assertEqual(
            result[0]["tag"],
            {
                "type": "fancy",
            },
        )

    def test_NodeRelationsUnusedElement(self):
        self._session_mock()

        result = self.api.NodeRelations(4295668179)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], self.api_base + "/api/0.6/node/4295668179/relations")

        self.assertEqual(len(result), 0)
        self.assertIsInstance(result, list)

    def test_NodesGet(self):
        self._session_mock()

        result = self.api.NodesGet([123, 345])

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], f"{self.api_base}/api/0.6/nodes?nodes=123,345")

        self.assertEqual(len(result), 2)
        self.assertEqual(
            result[123],
            {
                "id": 123,
                "changeset": 15293,
                "uid": 605,
                "timestamp": datetime.datetime(2012, 4, 18, 11, 14, 26),
                "lat": 51.8753146,
                "lon": -1.4857118,
                "visible": True,
                "version": 8,
                "user": "freundchen",
                "tag": {"amenity": "school", "foo": "bar", "name": "Berolina & Schule"},
            },
        )
        self.assertEqual(
            result[345],
            {
                "id": 345,
                "changeset": 244,
                "timestamp": datetime.datetime(2009, 9, 12, 3, 22, 59),
                "uid": 1,
                "visible": False,
                "version": 2,
                "user": "guggis",
                "tag": {},
            },
        )



import osmapi
import logging
from pprint import pformat
import urllib3

log = logging.getLogger(__name__)

loglevel = logging.DEBUG
logging.basicConfig(
    format="%(asctime)s %(name)s %(levelname)-8s %(message)s",
    level=loglevel,
    datefmt="%Y-%m-%d %H:%M:%S",
)
logging.captureWarnings(True)

# shut up DEBUG messages of specific loggers
logging.getLogger(osmapi.dom.__name__).setLevel(logging.INFO)
logging.getLogger(urllib3.__name__).setLevel(logging.INFO)


api = osmapi.OsmApi(api="https://api06.dev.openstreetmap.org")
node1 = api.NodeGet("1111")
log.debug(pformat(node1))



from dotenv import load_dotenv, find_dotenv
from oauthcli import OpenStreetMapDevAuth
from oauthlib.oauth2.rfc6749.errors import OAuth2Error
import logging
import os
import osmapi
import requests
import subprocess
import sys
import urllib3


load_dotenv(find_dotenv())

# logging setup
log = logging.getLogger(__name__)
loglevel = logging.DEBUG
logging.basicConfig(
    format="%(asctime)s %(name)s %(levelname)-8s %(message)s",
    level=loglevel,
    datefmt="%Y-%m-%d %H:%M:%S",
)
logging.captureWarnings(True)

# shut up DEBUG messages of specific loggers
logging.getLogger(osmapi.dom.__name__).setLevel(logging.INFO)
logging.getLogger(urllib3.__name__).setLevel(logging.INFO)


def clear_screen():
    # check and make call for specific operating system
    _ = subprocess.call("clear" if os.name == "posix" else "cls")


# The error handling with osmapi is very easy, simply catch the
# exception for the specific case you want to handle.
# - All osmapi excepctions are child classes of osmapi.OsmApiError
# - Errors that result from the communication with the OSM server osmapi.ApiError
# - There are a number of subclasses to differantiate the different errors
# - catch more specific errors first, then use more general error classes

# Upload data to OSM without a changeset
log.debug("Try to write data to OSM without a changeset")
api = osmapi.OsmApi(api="https://api06.dev.openstreetmap.org")
try:
    node1 = api.NodeCreate({"lon": 1, "lat": 1, "tag": {}})
except osmapi.NoChangesetOpenError as e:
    log.exception(e)
    log.debug("There is no open changeset")
input("Press Enter to continue...")
clear_screen()


# wrong server: ConnectionError
log.debug("Connect to wrong server...")
api = osmapi.OsmApi(api="https://invalid.server.name")
try:
    api.ChangesetGet(123)
except osmapi.ConnectionApiError as e:
    log.exception(e)
    log.debug("Error connecting to server")
input("Press Enter to continue...")
clear_screen()


# changeset not found: ElementNotFoundApiError
log.debug("Request non-existent changeset id...")
api = osmapi.OsmApi(api="https://api06.dev.openstreetmap.org")
try:
    api.ChangesetGet(111111111111)
except osmapi.ElementNotFoundApiError as e:
    log.exception(e)
    log.debug("Changeset not found")
input("Press Enter to continue...")
clear_screen()


# unauthorized request
log.debug("Try to add data with wrong authorization")
try:
    s = requests.Session()
    s.auth = ("user", "pass")
    api = osmapi.OsmApi(api="https://api06.dev.openstreetmap.org", session=s)
    with api.Changeset({"comment": "My first test"}) as changeset_id:
        node1 = api.NodeCreate({"lon": 1, "lat": 1, "tag": {}})
except osmapi.UnauthorizedApiError as e:
    log.exception(e)
    log.debug("Unauthorized to make this request")
input("Press Enter to continue...")
clear_screen()

# request without auhorization
log.debug("Try to add data without authorization")
try:
    api = osmapi.OsmApi(api="https://api06.dev.openstreetmap.org")
    with api.Changeset({"comment": "My first test"}) as changeset_id:
        node1 = api.NodeCreate({"lon": 1, "lat": 1, "tag": {}})
except osmapi.UsernamePasswordMissingError as e:
    log.exception(e)
    log.debug("Username/Password or authorization missing")
input("Press Enter to continue...")
clear_screen()


# a more or less complete "real-life" example
client_id = os.getenv("OSM_OAUTH_CLIENT_ID")
client_secret = os.getenv("OSM_OAUTH_CLIENT_SECRET")

try:
    auth = OpenStreetMapDevAuth(
        client_id, client_secret, ["write_api", "write_notes"]
    ).auth_code()
except OAuth2Error as e:
    log.exception(e)
    log.debug("An OAuth2 error occured")
    sys.exit(1)

try:
    api = osmapi.OsmApi(api="https://api06.dev.openstreetmap.org", session=auth.session)
    with api.Changeset({"comment": "My first test"}) as changeset_id:
        log.debug(f"Part of Changeset {changeset_id}")
        node1 = api.NodeCreate({"lon": 1, "lat": 1, "tag": {}})
        log.debug(node1)

    # get all the info from the closed changeset
    changeset = api.ChangesetGet(changeset_id)
    log.debug(changeset)
    exit_code = 0
except osmapi.ConnectionApiError as e:
    log.debug(f"Connection error: {str(e)}")
    exit_code = 1
    # display error for user, try again?
except osmapi.ElementNotFoundApiError as e:
    log.debug(f"Changeset not found: {str(e)}")
    exit_code = 1
except osmapi.ApiError as e:
    log.debug(f"Error on the API side: {str(e)}")
    exit_code = 1
except osmapi.OsmApiError as e:
    log.debug(f"Some other error: {str(e)}")
    exit_code = 1
finally:
    log.debug(f"Exit code: {exit_code}")
    sys.exit(exit_code)



import osmapi
import xmltodict
import datetime
import pytest
from responses import GET, PUT, POST
import requests


def xmltosorteddict(xml):
    xml_dict = xmltodict.parse(xml, dict_constructor=dict)
    return xml_dict


def test_Changeset_contextmanager(auth_api, add_response):
    # Setup mock
    resp = add_response(PUT, "/changeset/create", filename="test_Changeset_create.xml")
    resp = add_response(PUT, "/node/create", filename="test_Changeset_create_node.xml")
    resp = add_response(
        PUT, "/changeset/1414/close", filename="test_Changeset_close.xml"
    )

    test_node = {
        "lat": 47.123,
        "lon": 8.555,
        "tag": {"amenity": "place_of_worship", "religion": "pastafarian"},
    }

    # use context manager
    with auth_api.Changeset() as changeset_id:
        assert changeset_id == 1414

        # add test node
        node = auth_api.NodeCreate(test_node)
        assert node["id"] == 7272

    # check requests
    assert len(resp.calls) == 3


def test_ChangesetGet(api, add_response):
    # Setup mock
    add_response(GET, "/changeset/123")

    # Call
    result = api.ChangesetGet(123)

    test_changeset = {
        "id": 123,
        "closed_at": datetime.datetime(2009, 9, 7, 22, 57, 37),
        "created_at": datetime.datetime(2009, 9, 7, 21, 57, 36),
        "max_lat": "52.4710193",
        "max_lon": "-1.4831815",
        "min_lat": "45.9667901",
        "min_lon": "-1.4998534",
        "open": False,
        "user": "randomjunk",
        "uid": 3,
        "tag": {
            "comment": "correct node bug",
            "created_by": "Potlatch 1.2a",
        },
    }
    assert result == test_changeset


def test_ChangesetGet_with_connection_error(api, add_response):
    # Setup mock
    add_response(
        GET,
        "/changeset/123",
        body=requests.exceptions.ConnectionError("Connection aborted."),
    ),

    # Call
    with pytest.raises(osmapi.ConnectionApiError) as execinfo:
        api.ChangesetGet(123)
    assert (
        str(execinfo.value)
        == "Request failed: 0 - Connection error: Connection aborted. - "
    )


def test_ChangesetGet_with_timeout(api, add_response):
    # Setup mock
    add_response(GET, "/changeset/123", body=requests.exceptions.Timeout())

    # Call
    with pytest.raises(osmapi.TimeoutApiError) as execinfo:
        api.ChangesetGet(123)
    assert (
        str(execinfo.value) == "Request failed: 0 - Request timed out (timeout=30) - "
    )


def test_ChangesetUpdate(auth_api, add_response):
    # Setup mock
    resp = add_response(PUT, "/changeset/create", filename="test_ChangesetCreate.xml")
    resp = add_response(PUT, "/changeset/4321", filename="test_ChangesetUpdate.xml")

    # Call
    result = auth_api.ChangesetCreate()
    assert result == 4321

    result = auth_api.ChangesetUpdate({"test": "foobar"})
    changeset_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osm version="0.6" generator="osmapi/4.3.0">\n'
        b'  <changeset visible="true">\n'
        b'    <tag k="test" v="foobar"/>\n'
        b'    <tag k="created_by" v="osmapi/4.3.0"/>\n'
        b"  </changeset>\n"
        b"</osm>\n"
    )
    assert xmltosorteddict(resp.calls[1].request.body) == changeset_xml
    assert result == 4321


def test_ChangesetUpdate_with_created_by(auth_api, add_response):
    # Setup mock
    resp = add_response(PUT, "/changeset/create", filename="test_ChangesetCreate.xml")
    resp = add_response(PUT, "/changeset/4321", filename="test_ChangesetUpdate.xml")

    # Call
    result = auth_api.ChangesetCreate()
    assert result == 4321

    result = auth_api.ChangesetUpdate({"test": "foobar", "created_by": "MyTestOSMApp"})
    changeset_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osm version="0.6" generator="osmapi/4.3.0">\n'
        b'  <changeset visible="true">\n'
        b'    <tag k="test" v="foobar"/>\n'
        b'    <tag k="created_by" v="MyTestOSMApp"/>\n'
        b"  </changeset>\n"
        b"</osm>\n"
    )
    assert xmltosorteddict(resp.calls[1].request.body) == changeset_xml
    assert result == 4321


def test_ChangesetUpdate_wo_changeset(auth_api):
    with pytest.raises(osmapi.NoChangesetOpenError) as execinfo:
        auth_api.ChangesetUpdate({"test": "foobar"})
    assert str(execinfo.value) == "No changeset currently opened"


def test_ChangesetCreate(auth_api, add_response):
    resp = add_response(PUT, "/changeset/create")
    result = auth_api.ChangesetCreate({"foobar": "A new test changeset"})
    assert result == 4321

    changeset_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osm version="0.6" generator="osmapi/4.3.0">\n'
        b'  <changeset visible="true">\n'
        b'    <tag k="foobar" v="A new test changeset"/>\n'
        b'    <tag k="created_by" v="osmapi/4.3.0"/>\n'
        b"  </changeset>\n"
        b"</osm>\n"
    )
    assert xmltosorteddict(resp.calls[0].request.body) == changeset_xml


def test_ChangesetCreate_with_created_by(auth_api, add_response):
    resp = add_response(PUT, "/changeset/create")

    result = auth_api.ChangesetCreate(
        {
            "foobar": "A new test changeset",
            "created_by": "CoolTestApp",
        }
    )
    assert result == 1234

    changeset_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osm version="0.6" generator="osmapi/4.3.0">\n'
        b'  <changeset visible="true">\n'
        b'    <tag k="foobar" v="A new test changeset"/>\n'
        b'    <tag k="created_by" v="CoolTestApp"/>\n'
        b"  </changeset>\n"
        b"</osm>\n"
    )
    assert xmltosorteddict(resp.calls[0].request.body) == changeset_xml


def test_ChangesetCreate_with_open_changeset(auth_api, add_response):
    add_response(PUT, "/changeset/create")

    auth_api.ChangesetCreate(
        {
            "test": "an already open changeset",
        }
    )

    with pytest.raises(osmapi.ChangesetAlreadyOpenError) as execinfo:
        auth_api.ChangesetCreate({"test": "foobar"})
    assert str(execinfo.value) == "Changeset already opened"


def test_ChangesetCreate_with_prod_api_and_test_comment(prod_api):
    with pytest.raises(osmapi.OsmApiError) as execinfo:
        prod_api.ChangesetCreate(
            {
                "comment": "My first test",
            }
        )
    assert (
        str(execinfo.value) == "DO NOT CREATE test changesets on the production server"
    )


def test_ChangesetClose(auth_api, add_response):
    # setup mock
    resp = add_response(PUT, "/changeset/create", filename="test_Changeset_create.xml")
    resp = add_response(PUT, "/changeset/1414/close")

    # Call
    auth_api.ChangesetCreate()
    auth_api.ChangesetClose()

    assert "/api/0.6/changeset/1414/close" in resp.calls[1].request.url


def test_ChangesetClose_with_no_changeset(auth_api):
    with pytest.raises(osmapi.NoChangesetOpenError) as execinfo:
        auth_api.ChangesetClose()
    assert str(execinfo.value) == "No changeset currently opened"


def test_ChangesetUpload_create_node(auth_api, add_response):
    # Setup
    resp = add_response(PUT, "/changeset/create", body="4444")
    resp = add_response(POST, "/changeset/4444/upload")

    changesdata = [
        {
            "type": "node",
            "action": "create",
            "data": [
                {
                    "lat": 47.123,
                    "lon": 8.555,
                    "tag": {"amenity": "place_of_worship", "religion": "pastafarian"},
                },
                {
                    "lat": 47.125,
                    "lon": 8.557,
                    "tag": {"amenity": "place_of_worship", "religion": "pastafarian"},
                },
            ],
        }
    ]

    upload_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osmChange version="0.6" generator="osmapi/4.3.0">\n'
        b"<create>\n"
        b'  <node lat="47.123" lon="8.555" visible="true" '
        b'changeset="4444">\n'
        b'    <tag k="amenity" v="place_of_worship"/>\n'
        b'    <tag k="religion" v="pastafarian"/>\n'
        b"  </node>\n"
        b'  <node lat="47.125" lon="8.557" visible="true" '
        b'changeset="4444">\n'
        b'    <tag k="amenity" v="place_of_worship"/>\n'
        b'    <tag k="religion" v="pastafarian"/>\n'
        b"  </node>\n"
        b"</create>\n"
        b"</osmChange>"
    )

    # Call
    auth_api.ChangesetCreate()
    result = auth_api.ChangesetUpload(changesdata)

    # Assert
    assert xmltosorteddict(resp.calls[1].request.body) == upload_xml
    assert result[0]["type"] == changesdata[0]["type"]
    assert result[0]["action"] == changesdata[0]["action"]

    data = result[0]["data"]
    assert data[0]["lat"] == changesdata[0]["data"][0]["lat"]
    assert data[0]["lon"] == changesdata[0]["data"][0]["lon"]
    assert data[0]["tag"] == changesdata[0]["data"][0]["tag"]
    assert data[0]["id"] == 4295832900
    assert result[0]["data"][0]["version"] == 1


def test_ChangesetUpload_modify_way(auth_api, add_response):
    # setup mock
    resp = add_response(PUT, "/changeset/create", body="4444")
    resp = add_response(POST, "/changeset/4444/upload")

    changesdata = [
        {
            "type": "way",
            "action": "modify",
            "data": [
                {
                    "id": 4294967296,
                    "version": 2,
                    "nd": [
                        4295832773,
                        4295832773,
                        4294967304,
                        4294967303,
                        4294967300,
                        4608751,
                        4294967305,
                        4294967302,
                        8548430,
                        4294967296,
                        4294967301,
                        4294967298,
                        4294967306,
                        7855737,
                        4294967297,
                        4294967299,
                    ],
                    "tag": {"highway": "secondary", "name": "Stansted Road"},
                }
            ],
        }
    ]

    upload_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osmChange version="0.6" generator="osmapi/4.3.0">\n'
        b"<modify>\n"
        b'  <way id="4294967296" version="2" visible="true" '
        b'changeset="4444">\n'
        b'    <tag k="highway" v="secondary"/>\n'
        b'    <tag k="name" v="Stansted Road"/>\n'
        b'    <nd ref="4295832773"/>\n'
        b'    <nd ref="4295832773"/>\n'
        b'    <nd ref="4294967304"/>\n'
        b'    <nd ref="4294967303"/>\n'
        b'    <nd ref="4294967300"/>\n'
        b'    <nd ref="4608751"/>\n'
        b'    <nd ref="4294967305"/>\n'
        b'    <nd ref="4294967302"/>\n'
        b'    <nd ref="8548430"/>\n'
        b'    <nd ref="4294967296"/>\n'
        b'    <nd ref="4294967301"/>\n'
        b'    <nd ref="4294967298"/>\n'
        b'    <nd ref="4294967306"/>\n'
        b'    <nd ref="7855737"/>\n'
        b'    <nd ref="4294967297"/>\n'
        b'    <nd ref="4294967299"/>\n'
        b"  </way>\n"
        b"</modify>\n"
        b"</osmChange>"
    )

    # Call
    auth_api.ChangesetCreate()
    result = auth_api.ChangesetUpload(changesdata)
    # Assert
    assert xmltosorteddict(resp.calls[1].request.body) == upload_xml

    assert result[0]["type"] == changesdata[0]["type"]
    assert result[0]["action"] == changesdata[0]["action"]

    data = result[0]["data"][0]
    print(data)
    assert data["nd"] == changesdata[0]["data"][0]["nd"]
    assert data["tag"] == changesdata[0]["data"][0]["tag"]
    assert data["id"] == 4294967296
    assert data["version"] == 3


def test_ChangesetUpload_delete_relation(auth_api, add_response):
    # setup mock
    resp = add_response(PUT, "/changeset/create", body="4444")
    resp = add_response(POST, "/changeset/4444/upload")

    changesdata = [
        {
            "type": "relation",
            "action": "delete",
            "data": [
                {
                    "id": 676,
                    "version": 2,
                    "member": [
                        {"ref": 4799, "role": "outer", "type": "way"},
                        {"ref": 9391, "role": "outer", "type": "way"},
                    ],
                    "tag": {
                        "admin_level": "9",
                        "boundary": "administrative",
                        "type": "multipolygon",
                    },
                }
            ],
        }
    ]

    upload_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osmChange version="0.6" generator="osmapi/4.3.0">\n'
        b"<delete>\n"
        b'  <relation id="676" version="2" visible="true" '
        b'changeset="4444">\n'
        b'    <tag k="admin_level" v="9"/>\n'
        b'    <tag k="boundary" v="administrative"/>\n'
        b'    <tag k="type" v="multipolygon"/>\n'
        b'    <member type="way" ref="4799" role="outer"/>\n'
        b'    <member type="way" ref="9391" role="outer"/>\n'
        b"  </relation>\n"
        b"</delete>\n"
        b"</osmChange>"
    )

    # Call
    auth_api.ChangesetCreate()
    result = auth_api.ChangesetUpload(changesdata)

    # Assert
    assert xmltosorteddict(resp.calls[1].request.body) == upload_xml
    assert result[0]["type"] == changesdata[0]["type"]
    assert result[0]["action"] == changesdata[0]["action"]

    data = result[0]["data"][0]
    assert data["member"], changesdata[0]["data"][0]["member"]
    assert data["tag"] == changesdata[0]["data"][0]["tag"]
    assert data["id"] == 676
    assert "version" not in data


def test_ChangesetUpload_invalid_response(auth_api, add_response):
    # setup mock
    add_response(PUT, "/changeset/create", body="4444")
    add_response(POST, "/changeset/4444/upload", body="4444")

    changesdata = [
        {
            "type": "relation",
            "action": "delete",
            "data": [
                {
                    "id": 676,
                    "version": 2,
                    "member": [
                        {"ref": 4799, "role": "outer", "type": "way"},
                        {"ref": 9391, "role": "outer", "type": "way"},
                    ],
                    "tag": {
                        "admin_level": "9",
                        "boundary": "administrative",
                        "type": "multipolygon",
                    },
                }
            ],
        }
    ]

    # Call + assert
    auth_api.ChangesetCreate()
    with pytest.raises(osmapi.XmlResponseInvalidError) as execinfo:
        auth_api.ChangesetUpload(changesdata)
    assert "The XML response from the OSM API is invalid" in str(execinfo.value)


def test_ChangesetUpload_no_auth(api):
    changesdata = [
        {
            "type": "node",
            "action": "create",
            "data": [
                {
                    "lat": 47.123,
                    "lon": 8.555,
                    "tag": {"amenity": "place_of_worship", "religion": "pastafarian"},
                }
            ],
        }
    ]

    with pytest.raises(osmapi.UsernamePasswordMissingError) as execinfo:
        api.ChangesetUpload(changesdata)
    assert str(execinfo.value) == "Username/Password missing"


def test_ChangesetDownload(api, add_response):
    # Setup mock
    add_response(GET, "/changeset/23123/download")

    # Call
    result = api.ChangesetDownload(23123)

    # Assertion
    assert len(result) == 16
    assert result[1] == (
        {
            "action": "create",
            "type": "node",
            "data": {
                "changeset": 23123,
                "id": 4295668171,
                "lat": 46.4909781,
                "lon": 11.2743295,
                "tag": {"highway": "traffic_signals"},
                "timestamp": datetime.datetime(2013, 5, 14, 10, 33, 4),
                "uid": 1178,
                "user": "tyrTester06",
                "version": 1,
                "visible": True,
            },
        }
    )


def test_ChangesetDownload_invalid_response(api, add_response):
    add_response(GET, "/changeset/23123/download")
    with pytest.raises(osmapi.XmlResponseInvalidError) as execinfo:
        api.ChangesetDownload(23123)
    assert "The XML response from the OSM API is invalid" in str(execinfo.value)


def test_ChangesetDownloadContainingUnicode(api, add_response):
    add_response(GET, "/changeset/37393499/download")

    # This changeset contains unicode tag values
    # Note that the fixture data has been reduced from the
    # original from openstreetmap.org
    result = api.ChangesetDownload(37393499)

    assert len(result) == 2
    assert result[1] == (
        {
            "action": "create",
            "type": "way",
            "data": {
                "changeset": 37393499,
                "id": 399491497,
                "nd": [4022271571, 4022271567, 4022271565],
                "tag": {
                    "highway": "service",
                    # UTF-8 encoded 'LATIN SMALL LETTER O WITH STROKE'
                    # Aka. 0xf8 in latin-1/ISO 8859-1             Emoji: 😀
                    "name": b"S\xc3\xb8nderskovvej".decode("utf-8") + " \U0001f600",
                    "service": "driveway",
                },
                "timestamp": datetime.datetime(2016, 2, 23, 16, 55, 35),
                "uid": 328556,
                "user": "InternationalUser",
                "version": 1,
                "visible": True,
            },
        }
    )


def test_ChangesetsGet(api, add_response):
    resp = add_response(GET, "/changesets")

    result = api.ChangesetsGet(only_closed=True, username="metaodi")

    assert resp.calls[0].request.params == {"display_name": "metaodi", "closed": "1"}
    assert len(result) == 10
    assert result[41417] == (
        {
            "closed_at": datetime.datetime(2014, 4, 29, 20, 25, 1),
            "created_at": datetime.datetime(2014, 4, 29, 20, 25, 1),
            "id": 41417,
            "max_lat": "58.8997467",
            "max_lon": "22.7364427",
            "min_lat": "58.8501594",
            "min_lon": "22.6984333",
            "open": False,
            "tag": {
                "comment": "Test delete of relation",
                "created_by": "iD 1.3.9",
                "imagery_used": "Bing",
            },
            "uid": 1841,
            "user": "metaodi",
        }
    )


def test_ChangesetGetWithComment(api, add_response):
    resp = add_response(GET, "/changeset/52924")

    result = api.ChangesetGet(52924, include_discussion=True)

    assert resp.calls[0].request.params == {"include_discussion": "true"}
    assert result == {
        "id": 52924,
        "closed_at": datetime.datetime(2015, 1, 1, 14, 54, 2),
        "created_at": datetime.datetime(2015, 1, 1, 14, 54, 1),
        "comments_count": 3,
        "max_lat": "58.3369242",
        "max_lon": "25.8829107",
        "min_lat": "58.336813",
        "min_lon": "25.8823273",
        "discussion": [
            {
                "date": datetime.datetime(2015, 1, 1, 18, 56, 48),
                "text": "test",
                "uid": 1841,
                "user": "metaodi",
            },
            {
                "date": datetime.datetime(2015, 1, 1, 18, 58, 3),
                "text": "another comment",
                "uid": 1841,
                "user": "metaodi",
            },
            {
                "date": datetime.datetime(2015, 1, 1, 19, 16, 5),
                "text": "hello",
                "uid": 1841,
                "user": "metaodi",
            },
        ],
        "open": False,
        "user": "metaodi",
        "uid": 1841,
        "tag": {
            "comment": "My test",
            "created_by": "osmapi/0.4.1",
        },
    }


def test_ChangesetGetWithoutDiscussion(api, add_response):
    resp = add_response(GET, "/changeset/52924")

    result = api.ChangesetGet(52924, include_discussion=False)

    assert resp.calls[0].request.params == {}
    assert result == {
        "id": 52924,
        "closed_at": datetime.datetime(2015, 1, 1, 14, 54, 2),
        "created_at": datetime.datetime(2015, 1, 1, 14, 54, 1),
        "max_lat": "58.3369242",
        "max_lon": "25.8829107",
        "min_lat": "58.336813",
        "min_lon": "25.8823273",
        "open": False,
        "user": "metaodi",
        "uid": 1841,
        "tag": {
            "comment": "My test",
            "created_by": "osmapi/0.4.1",
        },
    }


def test_ChangesetComment(auth_api, add_response):
    resp = add_response(POST, "/changeset/123/comment")

    result = auth_api.ChangesetComment(123, comment="test comment")

    assert resp.calls[0].request.body == "text=test+comment"
    assert result == {
        "id": 123,
        "closed_at": datetime.datetime(2009, 9, 7, 22, 57, 37),
        "created_at": datetime.datetime(2009, 9, 7, 21, 57, 36),
        "max_lat": "52.4710193",
        "max_lon": "-1.4831815",
        "min_lat": "45.9667901",
        "min_lon": "-1.4998534",
        "open": False,
        "user": "randomjunk",
        "uid": 3,
        "tag": {
            "comment": "correct node bug",
            "created_by": "Potlatch 1.2a",
        },
    }


def test_ChangesetComment_no_auth(api):
    with pytest.raises(osmapi.UsernamePasswordMissingError) as execinfo:
        api.ChangesetComment(123, comment="test comment")
    assert str(execinfo.value) == "Username/Password missing"


def test_ChangesetSubscribe(auth_api, add_response):
    add_response(POST, "/changeset/123/subscribe")

    result = auth_api.ChangesetSubscribe(123)

    assert result == {
        "id": 123,
        "closed_at": datetime.datetime(2009, 9, 7, 22, 57, 37),
        "created_at": datetime.datetime(2009, 9, 7, 21, 57, 36),
        "max_lat": "52.4710193",
        "max_lon": "-1.4831815",
        "min_lat": "45.9667901",
        "min_lon": "-1.4998534",
        "open": False,
        "user": "randomjunk",
        "uid": 3,
        "tag": {
            "comment": "correct node bug",
            "created_by": "Potlatch 1.2a",
        },
    }


def test_ChangesetSubscribeWhenAlreadySubscribed(auth_api, add_response):
    add_response(POST, "/changeset/52924/subscribe", status=409)

    with pytest.raises(osmapi.AlreadySubscribedApiError) as execinfo:
        auth_api.ChangesetSubscribe(52924)

    assert execinfo.value.payload == b"You are already subscribed to changeset 52924."
    assert execinfo.value.reason == "Conflict"
    assert execinfo.value.status == 409


def test_ChangesetSubscribe_no_auth(api):
    with pytest.raises(osmapi.UsernamePasswordMissingError) as execinfo:
        api.ChangesetSubscribe(45627)
    assert str(execinfo.value) == "Username/Password missing"


def test_ChangesetUnsubscribe(auth_api, add_response):
    add_response(POST, "/changeset/123/unsubscribe")

    result = auth_api.ChangesetUnsubscribe(123)

    assert result == {
        "id": 123,
        "closed_at": datetime.datetime(2009, 9, 7, 22, 57, 37),
        "created_at": datetime.datetime(2009, 9, 7, 21, 57, 36),
        "max_lat": "52.4710193",
        "max_lon": "-1.4831815",
        "min_lat": "45.9667901",
        "min_lon": "-1.4998534",
        "open": False,
        "user": "randomjunk",
        "uid": 3,
        "tag": {
            "comment": "correct node bug",
            "created_by": "Potlatch 1.2a",
        },
    }


def test_ChangesetUnsubscribeWhenNotSubscribed(auth_api, add_response):
    add_response(POST, "/changeset/52924/unsubscribe", status=404)

    with pytest.raises(osmapi.NotSubscribedApiError) as execinfo:
        auth_api.ChangesetUnsubscribe(52924)

    assert execinfo.value.payload == b"You are not subscribed to changeset 52924."
    assert execinfo.value.reason == "Not Found"
    assert execinfo.value.status == 404


def test_ChangesetUnsubscribe_no_auth(api):
    with pytest.raises(osmapi.UsernamePasswordMissingError) as execinfo:
        api.ChangesetUnsubscribe(45627)
    assert str(execinfo.value) == "Username/Password missing"



import osmapi
from pprint import pprint

api = osmapi.OsmApi(api="https://api06.dev.openstreetmap.org")

try:
    api.ChangesetGet(111111111111)
except osmapi.ApiError as e:
    print(f"Error: {e}")
    if e.status == 404:
        print("Changeset not found")

print("")
pprint(api.ChangesetGet(12345))



from . import osmapi_test
import osmapi
from unittest import mock
import os

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


class TestOsmApiHelper(osmapi_test.TestOsmApi):
    def setUp(self):
        super().setUp()
        self.setupMock()

    def setupMock(self, status=200):
        mock_response = mock.Mock()
        mock_response.status_code = status
        mock_response.reason = "test reason"
        mock_response.content = "test response"

        self.mock_session = mock.Mock()
        self.mock_session.request = mock.Mock(return_value=mock_response)
        self.mock_session.close = mock.Mock()
        self.mock_session.auth = ("testuser", "testpassword")

        self.api = osmapi.OsmApi(
            api=self.api_base,
            session=self.mock_session,
            username="testuser",
            password="testpassword",
        )

    def test_passwordfile_only(self):
        path = os.path.join(__location__, "fixtures", "passwordfile.txt")
        my_api = osmapi.OsmApi(passwordfile=path)
        self.assertEqual("testosm", my_api._username)
        self.assertEqual("testpass", my_api._password)

    def test_passwordfile_with_user(self):
        path = os.path.join(__location__, "fixtures", "passwordfile.txt")
        my_api = osmapi.OsmApi(username="testuser", passwordfile=path)
        self.assertEqual("testuser", my_api._username)
        self.assertEqual("testuserpass", my_api._password)

    def test_passwordfile_with_colon(self):
        path = os.path.join(__location__, "fixtures", "passwordfile_colon.txt")
        my_api = osmapi.OsmApi(username="testuser", passwordfile=path)
        self.assertEqual("testuser", my_api._username)
        self.assertEqual("test:userpass", my_api._password)

    def test_close_call(self):
        self.api.close()
        self.assertEqual(self.api._session._session.close.call_count, 1)

    def test_close_context_manager(self):
        with osmapi.OsmApi() as my_api:
            my_api._session.close = mock.Mock()
        self.assertEqual(my_api._session.close.call_count, 1)

    def test_http_request_get(self):
        response = self.api._session._http_request("GET", "/api/0.6/test", False, None)
        self.mock_session.request.assert_called_with(
            "GET", self.api_base + "/api/0.6/test", data=None, timeout=30
        )
        self.assertEqual(response, "test response")
        self.assertEqual(self.mock_session.request.call_count, 1)

    def test_http_request_put(self):
        data = "data"
        response = self.api._session._http_request(
            "PUT", "/api/0.6/testput", False, data
        )
        self.mock_session.request.assert_called_with(
            "PUT", self.api_base + "/api/0.6/testput", data="data", timeout=30
        )
        self.assertEqual(response, "test response")

    def test_http_request_delete(self):
        data = "delete data"
        response = self.api._session._http_request(
            "PUT", "/api/0.6/testdelete", False, data
        )
        self.mock_session.request.assert_called_with(
            "PUT", self.api_base + "/api/0.6/testdelete", data="delete data", timeout=30
        )
        self.assertEqual(response, "test response")

    def test_http_request_auth(self):
        response = self.api._session._http_request(
            "PUT", "/api/0.6/testauth", True, None
        )
        self.mock_session.request.assert_called_with(
            "PUT", self.api_base + "/api/0.6/testauth", data=None, timeout=30
        )
        self.assertEqual(self.mock_session.auth, ("testuser", "testpassword"))
        self.assertEqual(response, "test response")

    def test_http_request_410_response(self):
        self.setupMock(410)
        with self.assertRaises(osmapi.ElementDeletedApiError) as cm:
            self.api._session._http_request("GET", "/api/0.6/test410", False, None)
        self.assertEqual(cm.exception.status, 410)
        self.assertEqual(cm.exception.reason, "test reason")
        self.assertEqual(cm.exception.payload, "test response")

    def test_http_request_500_response(self):
        self.setupMock(500)
        with self.assertRaises(osmapi.ApiError) as cm:
            self.api._session._http_request(
                "GET", self.api_base + "/api/0.6/test500", False, None
            )
        self.assertEqual(cm.exception.status, 500)
        self.assertEqual(cm.exception.reason, "test reason")
        self.assertEqual(cm.exception.payload, "test response")



from . import osmapi_test
import osmapi
from unittest import mock
import datetime


class TestOsmApiDom(osmapi_test.TestOsmApi):
    def test_DomGetAttributes(self):
        mock_domelement = mock.Mock()
        mock_domelement.attributes = {
            "uid": "12345",
            "open": "false",
            "visible": "true",
            "lat": "47.1234",
            "date": "2021-12-10T21:28:03Z",
            "new_attribute": "Test 123",
        }

        result = osmapi.dom._DomGetAttributes(mock_domelement)

        self.assertIsInstance(result, dict)
        self.assertEqual(result["uid"], 12345)
        self.assertEqual(result["open"], False)
        self.assertEqual(result["visible"], True)
        self.assertEqual(result["lat"], 47.1234)
        self.assertEqual(result["date"], datetime.datetime(2021, 12, 10, 21, 28, 3))
        self.assertEqual(result["new_attribute"], "Test 123")

    def test_ParseDate(self):
        self.assertEqual(
            osmapi.dom._ParseDate("2021-02-25T09:49:33Z"),
            datetime.datetime(2021, 2, 25, 9, 49, 33),
        )
        self.assertEqual(
            osmapi.dom._ParseDate("2021-02-25 09:49:33 UTC"),
            datetime.datetime(2021, 2, 25, 9, 49, 33),
        )
        with self.assertLogs("osmapi.dom", level="DEBUG") as cm:
            self.assertEqual(osmapi.dom._ParseDate("2021-02-25"), "2021-02-25")
            self.assertEqual(osmapi.dom._ParseDate(""), "")
            self.assertIsNone(osmapi.dom._ParseDate(None))

            # test logging output
            self.assertEqual(
                cm.output,
                [
                    "DEBUG:osmapi.dom:2021-02-25 does not match %Y-%m-%d %H:%M:%S UTC",
                    "DEBUG:osmapi.dom:2021-02-25 does not match %Y-%m-%dT%H:%M:%SZ",
                    "DEBUG:osmapi.dom: does not match %Y-%m-%d %H:%M:%S UTC",
                    "DEBUG:osmapi.dom: does not match %Y-%m-%dT%H:%M:%SZ",
                    "DEBUG:osmapi.dom:None does not match %Y-%m-%d %H:%M:%S UTC",
                    "DEBUG:osmapi.dom:None does not match %Y-%m-%dT%H:%M:%SZ",
                ],
            )



from . import osmapi_test


class TestOsmApiNode(osmapi_test.TestOsmApi):
    def test_Capabilities(self):
        self._session_mock()

        result = self.api.Capabilities()
        self.assertEqual(
            result,
            {
                "area": {"maximum": 0.25},
                "changesets": {"maximum_elements": 50000.0},
                "status": {"api": "mocked", "database": "online", "gpx": "online"},
                "timeout": {"seconds": 300.0},
                "tracepoints": {"per_page": 5000.0},
                "version": {"maximum": 0.6, "minimum": 0.6},
                "waynodes": {"maximum": 2000.0},
            },
        )



from . import osmapi_test
import osmapi
from unittest import mock
import datetime


class TestOsmApiWay(osmapi_test.TestOsmApi):
    def test_WayGet(self):
        self._session_mock()

        result = self.api.WayGet(321)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], self.api_base + "/api/0.6/way/321")

        self.assertEqual(
            result,
            {
                "id": 321,
                "changeset": 298,
                "uid": 12,
                "timestamp": datetime.datetime(2009, 9, 14, 23, 23, 18),
                "visible": True,
                "version": 1,
                "user": "green525",
                "tag": {
                    "admin_level": "9",
                    "boundary": "administrative",
                },
                "nd": [
                    11949,
                    11950,
                    11951,
                    11952,
                    11953,
                    11954,
                    11955,
                    11956,
                    11957,
                    11958,
                    11959,
                    11960,
                    11961,
                    11962,
                    11963,
                    11964,
                    11949,
                ],
            },
        )

    def test_WayGet_with_version(self):
        self._session_mock()

        result = self.api.WayGet(4294967296, 2)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], f"{self.api_base}/api/0.6/way/4294967296/2")

        self.assertEqual(result["id"], 4294967296)
        self.assertEqual(result["changeset"], 41303)
        self.assertEqual(result["user"], "metaodi")

    def test_WayGet_nodata(self):
        self._session_mock()

        with self.assertRaises(osmapi.ResponseEmptyApiError):
            self.api.WayGet(321)

    def test_WayCreate(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=2222)
        self.api._CurrentChangesetId = 2222

        test_way = {
            "nd": [11949, 11950],
            "tag": {"highway": "unclassified", "name": "Osmapi Street"},
        }

        cs = self.api.ChangesetCreate({"comment": "This is a test way"})
        self.assertEqual(cs, 2222)

        result = self.api.WayCreate(test_way)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "PUT")
        self.assertEqual(args[1], self.api_base + "/api/0.6/way/create")

        self.assertEqual(result["id"], 5454)
        self.assertEqual(result["nd"], test_way["nd"])
        self.assertEqual(result["tag"], test_way["tag"])

    def test_WayCreate_existing_node(self):
        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=1111)
        self.api._CurrentChangesetId = 1111

        test_way = {
            "id": 456,
            "nd": [11949, 11950],
            "tag": {"highway": "unclassified", "name": "Osmapi Street"},
        }

        with self.assertRaisesRegex(
            osmapi.OsmTypeAlreadyExistsError, "This way already exists"
        ):
            self.api.WayCreate(test_way)

    def test_WayUpdate(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=2222)
        self.api._CurrentChangesetId = 2222

        test_way = {
            "id": 876,
            "nd": [11949, 11950],
            "tag": {"highway": "unclassified", "name": "Osmapi Street Update"},
        }

        cs = self.api.ChangesetCreate({"comment": "This is a test way"})
        self.assertEqual(cs, 2222)

        result = self.api.WayUpdate(test_way)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "PUT")
        self.assertEqual(args[1], self.api_base + "/api/0.6/way/876")

        self.assertEqual(result["id"], 876)
        self.assertEqual(result["version"], 7)
        self.assertEqual(result["nd"], test_way["nd"])
        self.assertEqual(result["tag"], test_way["tag"])

    def test_WayUpdatePreconditionFailed(self):
        self._session_mock(auth=True, status=412)

        self.api.ChangesetCreate = mock.Mock(return_value=1111)
        self.api._CurrentChangesetId = 1111

        test_way = {
            "id": 876,
            "nd": [11949, 11950],
            "tag": {"highway": "unclassified", "name": "Osmapi Street Update"},
        }

        self.api.ChangesetCreate({"comment": "This is a test dataset"})

        with self.assertRaises(osmapi.PreconditionFailedApiError) as cm:
            self.api.WayUpdate(test_way)

        self.assertEqual(cm.exception.status, 412)
        self.assertEqual(
            cm.exception.payload,
            (
                "Way 876 requires the nodes with id in (11950), "
                "which either do not exist, or are not visible."
            ),
        )

    def test_WayDelete(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(return_value=2222)
        self.api._CurrentChangesetId = 2222

        test_way = {"id": 876}

        cs = self.api.ChangesetCreate({"comment": "This is a test way delete"})
        self.assertEqual(cs, 2222)

        result = self.api.WayDelete(test_way)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "DELETE")
        self.assertEqual(args[1], self.api_base + "/api/0.6/way/876")
        self.assertEqual(result["id"], 876)
        self.assertEqual(result["version"], 8)

    def test_WayHistory(self):
        self._session_mock()

        result = self.api.WayHistory(4294967296)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], f"{self.api_base}/api/0.6/way/4294967296/history")

        self.assertEqual(len(result), 2)
        self.assertEqual(result[1]["id"], 4294967296)
        self.assertEqual(result[1]["version"], 1)
        self.assertEqual(
            result[1]["tag"],
            {
                "highway": "unclassified",
                "name": "Stansted Road",
            },
        )

    def test_WayRelations(self):
        self._session_mock()

        result = self.api.WayRelations(4295032193)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], f"{self.api_base}/api/0.6/way/4295032193/relations")

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]["id"], 4294968148)
        self.assertEqual(result[0]["changeset"], 23123)
        self.assertEqual(
            result[0]["member"][4],
            {
                "role": "",
                "ref": 4295032193,
                "type": "way",
            },
        )
        self.assertEqual(
            result[0]["tag"],
            {
                "type": "fancy",
            },
        )

    def test_WayRelationsUnusedElement(self):
        self._session_mock()

        result = self.api.WayRelations(4295032193)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], self.api_base + "/api/0.6/way/4295032193/relations")

        self.assertEqual(len(result), 0)
        self.assertIsInstance(result, list)

    def test_WayFull(self):
        self._session_mock()

        result = self.api.WayFull(321)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], self.api_base + "/api/0.6/way/321/full")

        self.assertEqual(len(result), 17)
        self.assertEqual(result[0]["data"]["id"], 11949)
        self.assertEqual(result[0]["data"]["changeset"], 298)
        self.assertEqual(result[0]["type"], "node")
        self.assertEqual(result[16]["data"]["id"], 321)
        self.assertEqual(result[16]["data"]["changeset"], 298)
        self.assertEqual(result[16]["type"], "way")

    def test_WayFull_invalid_response(self):
        self._session_mock()

        with self.assertRaises(osmapi.XmlResponseInvalidError):
            self.api.WayFull(321)

    def test_WaysGet(self):
        self._session_mock()

        result = self.api.WaysGet([456, 678])

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], "GET")
        self.assertEqual(args[1], f"{self.api_base}/api/0.6/ways?ways=456,678")

        self.assertEqual(len(result), 2)
        self.assertIs(type(result[456]), dict)
        self.assertIs(type(result[678]), dict)
        with self.assertRaises(KeyError):
            self.assertIs(type(result[123]), dict)



# -*- coding: utf-8 -*-

from codecs import open
from setuptools import setup, find_packages
import re

with open("osmapi/__init__.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

if not version:
    raise RuntimeError("Cannot find version information")

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="osmapi",
    packages=find_packages(),
    version=version,
    install_requires=["requests"],
    python_requires=">=3.8",
    description="Python wrapper for the OSM API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Etienne Chové",
    author_email="chove@crans.org",
    maintainer="Stefan Oderbolz",
    maintainer_email="odi@metaodi.ch",
    url="https://github.com/metaodi/osmapi",
    download_url=f"https://github.com/metaodi/osmapi/archive/v{version}.zip",
    keywords=["openstreetmap", "osm", "api"],
    license="GPLv3",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Software Development :: Libraries",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



# install cli-oauth2 for requests:  pip install cli-oauth2
from oauthcli import OpenStreetMapDevAuth
import osmapi
from dotenv import load_dotenv, find_dotenv
import os

load_dotenv(find_dotenv())

# load secrets for OAuth
client_id = os.getenv("OSM_OAUTH_CLIENT_ID")
client_secret = os.getenv("OSM_OAUTH_CLIENT_SECRET")

auth = OpenStreetMapDevAuth(
    client_id, client_secret, ["write_api", "write_notes"]
).auth_code()


# Use a normal timeout (30s is the default value)
normal_timeout_api = osmapi.OsmApi(
    api="https://api06.dev.openstreetmap.org", session=auth.session, timeout=30
)
changeset_id = normal_timeout_api.ChangesetCreate({"comment": "My first test"})
print(f"Create new changeset {changeset_id}")

# Deliberately using a very small timeout to show what happens when a timeout occurs
low_timeout_api = osmapi.OsmApi(
    api="https://api06.dev.openstreetmap.org", session=auth.session, timeout=0.00001
)
try:
    changeset_id = low_timeout_api.ChangesetCreate({"comment": "My first test"})
    print(f"Create new changeset {changeset_id}")
except osmapi.errors.TimeoutApiError as e:
    print(f"Timeout error occured: {str(e)}")
